@@ -4,7 +4,7 @@ use strict;
 use warnings;
 use Mail::Milter::Authentication::Pragmas;
 # ABSTRACT: Handler superclass
-our $VERSION = '2.20200625.2'; # VERSION
+our $VERSION = '2.20200930.2'; # VERSION
 use Mail::Milter::Authentication::Exception;
 use Mail::Milter::Authentication::Resolver;
 use Date::Format qw{ time2str };
@@ -13,6 +13,7 @@ use MIME::Base64;
 use Mail::SPF;
 use Net::DNS::Resolver;
 use Net::IP;
+use Sereal qw{encode_sereal decode_sereal};
 use Sys::Hostname;
 use Time::HiRes qw{ ualarm gettimeofday };
 
@@ -165,6 +166,41 @@ sub register_metrics {
 }
 
 
+sub top_dequeue_callback {
+    my ( $self ) = @_;
+
+    $self->status('dequeue');
+    $self->set_symbol('C','i','DEQUEUE.'.substr( uc md5_hex( "Authentication Milter Client $PID " . time() . rand(100) ) , -11 ));
+    $self->dbgout( 'CALLBACK', 'Dequeue', LOG_DEBUG );
+    my $config = $self->config();
+    eval {
+        local $SIG{'ALRM'} = sub{ die Mail::Milter::Authentication::Exception->new({ 'Type' => 'Timeout', 'Text' => 'Dequeue callback timeout' }) };
+        if ( my $timeout = $self->get_type_timeout( 'dequeue' ) ) {
+            $self->set_alarm( $timeout );
+        }
+        my $callbacks = $self->get_callbacks( 'dequeue' );
+        foreach my $handler ( @$callbacks ) {
+            $self->dbgout( 'CALLBACK', 'Dequeue ' . $handler, LOG_DEBUG );
+            my $start_time = $self->get_microseconds();
+            $self->get_handler($handler)->dequeue_callback();
+            $self->dbgoutwrite();
+            $self->metric_count( 'time_microseconds_total', { 'callback' => 'dequeue', 'handler' => $handler }, $self->get_microseconds_since( $start_time ) );
+        }
+        $self->set_alarm(0);
+    };
+    if ( my $error = $@ ) {
+        if ( my $type = $self->is_exception_type( $error ) ) {
+            $self->metric_count( 'callback_error_total', { 'stage' => 'dequeue', 'type' => $type } );
+        }
+        else {
+            $self->metric_count( 'callback_error_total', { 'stage' => 'dequeue' } );
+        }
+    }
+    $self->dbgoutwrite();
+    $self->status('postdequeue');
+}
+
+
 sub top_setup_callback {
 
     my ( $self ) = @_;
@@ -235,11 +271,11 @@ sub set_handler_alarm {
     my $remaining = $self->get_time_remaining();
     if ( $remaining < $microseconds ) {
         # This should already be set of course, but for clarity...
-        $self->dbgout( 'Handler tmeout set (remaining used)', $remaining, LOG_DEBUG );
+        $self->dbgout( 'Handler timeout set (remaining used)', $remaining, LOG_DEBUG );
         ualarm( $remaining );
     }
     else {
-        $self->dbgout( 'Handler tmeout set', $microseconds, LOG_DEBUG );
+        $self->dbgout( 'Handler timeout set', $microseconds, LOG_DEBUG );
         ualarm( $microseconds );
     }
 }
@@ -530,7 +566,7 @@ sub top_envfrom_callback {
 
     # On MAILFROM
     #...
-    my ( $self, $env_from ) = @_;
+    my ( $self, $env_from, @params ) = @_;
     $self->status('envfrom');
     $self->dbgout( 'CALLBACK', 'EnvFrom', LOG_DEBUG );
     $self->set_return( $self->smfis_continue() );
@@ -552,7 +588,7 @@ sub top_envfrom_callback {
         foreach my $handler ( @$callbacks ) {
             $self->dbgout( 'CALLBACK', 'EnvFrom ' . $handler, LOG_DEBUG );
             my $start_time = $self->get_microseconds();
-            eval { $self->get_handler($handler)->envfrom_callback($env_from); };
+            eval { $self->get_handler($handler)->envfrom_callback($env_from, @params); };
             if ( my $error = $@ ) {
                 $self->handle_exception( $error );
                 $self->exit_on_close( 'Env From callback error ' . $error );
@@ -584,7 +620,7 @@ sub top_envrcpt_callback {
 
     # On RCPTTO
     #...
-    my ( $self, $env_to ) = @_;
+    my ( $self, $env_to, @params ) = @_;
     $self->status('envrcpt');
     $self->dbgout( 'CALLBACK', 'EnvRcpt', LOG_DEBUG );
     $self->set_return( $self->smfis_continue() );
@@ -600,7 +636,7 @@ sub top_envrcpt_callback {
         foreach my $handler ( @$callbacks ) {
             $self->dbgout( 'CALLBACK', 'EnvRcpt ' . $handler, LOG_DEBUG );
             my $start_time = $self->get_microseconds();
-            eval{ $self->get_handler($handler)->envrcpt_callback($env_to); };
+            eval{ $self->get_handler($handler)->envrcpt_callback($env_to, @params); };
             if ( my $error = $@ ) {
                 $self->handle_exception( $error );
                 $self->exit_on_close( 'Env Rcpt callback error ' . $error );
@@ -1376,6 +1412,59 @@ sub tempfail_on_error {
 
 # Common calls into other Handlers
 
+sub _dequeue_dir($self) {
+    my $config = $self->config();
+    my $dir = $config->{spool_dir}.'/dequeue';
+    mkdir $dir if ! -d $dir;
+    return $dir;
+}
+
+
+{
+    my $queue_index = 1;
+    sub add_dequeue($self,$key,$data) {
+        my $dir = $self->_dequeue_dir;
+        my $fullpath;
+        my $timestamp = join( '.',gettimeofday);
+        my $filename = join( '.',$key,$PID,$timestamp,$queue_index++,'dequeue');
+        $fullpath = "$dir/$filename";
+        my $serialised_data = encode_sereal($data);
+        write_file($fullpath,{atomic=>1},$serialised_data);
+    }
+}
+
+
+sub get_dequeue_list($self,$key) {
+    my @dequeue_list;
+    my $dir = $self->_dequeue_dir;
+    opendir(my $dh, $dir) || die "Failed to open dequeue directory: $!";
+    while (my $file = readdir $dh) {
+        push @dequeue_list, $file if $file =~ /^$key\..*\.dequeue$/;
+    }
+    closedir $dh;
+    return \@dequeue_list;
+}
+
+
+sub get_dequeue($self,$id) {
+    my $dir = $self->_dequeue_dir;
+    my $filepath = join('/',$dir,$id);
+    return if ! -e $filepath;
+    return if ! -f $filepath;
+    my $serialized = scalar read_file($filepath);
+    my $data = decode_sereal($serialized);
+    return $data;
+}
+
+
+sub delete_dequeue($self,$id) {
+    my $dir = $self->_dequeue_dir;
+    my $filepath = join('/',$dir,$id);
+    return if ! -e $filepath;
+    return if ! -f $filepath;
+    unlink $filepath;
+}
+
 
 sub is_local_ip_address {
     my ($self) = @_;
@@ -1885,13 +1974,15 @@ sub add_headers {
             $header = $header_obj->as_string();
         }
 
+        $self->prepend_header( 'Authentication-Results', $header );
     }
-    else {
+    elsif ( !$config->{'hide_none'} ) {
         $header .= '; none';
+        $self->prepend_header( 'Authentication-Results', $header );
+    } else {
+        # the result is none and hide_none is set, so we do not add an AR header
     }
 
-    $self->prepend_header( 'Authentication-Results', $header );
-
     if ( my $reason = $self->get_quarantine_mail() ) {
         $self->prepend_header( 'X-Disposition-Quarantine', $reason );
     }
@@ -2044,7 +2135,7 @@ Mail::Milter::Authentication::Handler - Handler superclass
 
 =head1 VERSION
 
-version 2.20200625.2
+version 2.20200930.2
 
 =head1 DESCRIPTION
 
@@ -2333,6 +2424,10 @@ Manually check the current timeout, and throw if it has passed.
 
 =head1 CALLBACK METHODS
 
+=head2 I<top_dequeue_callback()>
+
+Top level handler for dequeue.
+
 =head2 I<top_setup_callback()>
 
 Top level handler for handler setup.
@@ -2401,6 +2496,28 @@ Called after the Authentication-Results header has been added, but before any ot
 
 =head1 HELPER METHODS
 
+=head2 I<add_dequeue($key,$data)>
+
+Write serialized $data into the queue for later dequeueing
+
+=head2 I<get_dequeue_list($key)>
+
+Return an ArrayRef of all queued items for $key
+
+This may be a list of filenames, or may be a list of some
+other ID, it should not be assumed that this value is
+useful outside of the dequeue methods.
+
+Used in get_dequeue_object and delete_dequeue_object
+
+=head2 I<get_dequeue($id)>
+
+Return a previously queued item
+
+=head2 I<delete_dequeue($id)>
+
+Delete a previously queued item
+
 =head2 I<is_local_ip_address()>
 
 Is the current connection from a local ip address?
